﻿SuperStrict

Framework Openb3dmax.B3dglgraphics
Import brl.jpgloader
Import brl.pngloader
Import brl.timer
Import maxgui.Drivers

' ------------------------------------------------------------------------------------------------
' Variables to change
' ------------------------------------------------------------------------------------------------

' bugcheck variables to set
Const CREATECOLUMNS:Int = True                    ' create the columns
Const CREATECEILING:Int = True                    ' create the ceiling
Const CREATELIGHTS:Int = True                     ' create the RGB lights
Const CREATEFIREBALL:Int = True                   ' create the fireball
Const APPLYSHADER:Int = True                      ' apply shader

' screen
Global FPS:Int = 60                               ' target FPS
Global FOV:Float = 115.0                          ' camera Field of View
Global SCREEN_WIDTH:Int = 1920                    ' width
Global SCREEN_HEIGHT:Int = 1080                   ' height

' shader variables
Global sh:String = "shader"                       ' shader filename prefix
Global levelscale:Float = 128.0                   ' overall level scaler (1...128)

' texture inputs
Global level:String = "test.stone"                ' textures subfolder
Global prefix:String = "test"                     ' filename prefix
Global texformat:String = "tga"                   ' file format extension
Global walltex:String = prefix' + "_wall"         ' walls
Global floortex:String = prefix' + "_floor"       ' floor
Global ceilingtex:String = prefix' + "_ceiling"   ' ceiling
Global columntex:String = prefix' + "_column"     ' columns

' Shader: Light
Global gamma:Float = 1.0                          ' overall screen gamma value
Global ambFactor:Float = 1.0                      ' mulitplicator of the ambient light value
Global AttA:Float = 5.0                           ' light falloff curve Var B
Global AttB:Float = 5.0                           ' light falloff curve Var B

' Shader: Fog
Global FogStart:Float = 4.0 * levelscale          ' Fog Range Start
Global FogRange:Float = 10.0 * levelscale         ' Fog Range End
Global FogDensity:Float = 0.25                    ' how "thick" the fog is
Global FogColor:Int[] = [8, 12, 20]               ' FogColor (and CameraCLScolor, Ambient Color)

' Texture Scaling [u,v,offsetx,offsety]
Global texturescale_multi:Float = 2.0
Global texturescale_wall:Float[] = [1.0, 1.0, 0.0, 0.0]
Global texturescale_floor:Float[] = [1.0, 1.0, 0.0, 0.0]
Global texturescale_ceiling:Float[] = [1.0, 1.0, 0.0, 0.0]
Global texturescale_columns:Float[] = [1.0, 1.0, 0.0, 0.0]
Global texturescale_sphere:Float[] = [2.0, 1.0, 0.0, 0.0]

' flags
Global DM:Int = False    ' Flag: Depthmap
Global FG:Int = True     ' Flag: Fog
Global PB:Int = True     ' Flag: PBR Shading
Global TM:Int = 5        ' Flag: Tonemapping Model
Global TL:Int = False    ' Flag: Torchlight
Global WF:Int = False    ' Flag: Wireframe
Global VS:Int = True     ' Flag: VSync

Global LR1:Float = 10.0 ' Lightrange (Torch)
Global LR2:Float = 4.0  ' Lightrange (Red Light)
Global LR3:Float = 4.0  ' Lightrange (Green Light)
Global LR4:Float = 4.0  ' Lightrange (Blue Light)
Global LRF:Float = 8.0  ' Lightrange (Fireball)


' ------------------------------------------------------------------------------------------------
' Other Variables
' ------------------------------------------------------------------------------------------------

' movement
Global m1:Float
Global m2:Float
Global m3:Float

' user variables
Global SCREENSHOT:Int

' text strings
Global s:String[] = ["off", "on"]
Global t:String[] = ["none", "Filmic", "Simple", "Exposure", "PBR", "Uncharted2", "S-Curve"]

Global keyAL:Int, keyNM:Int, keyRO:Int, keyME:Int, keyHM:Int, keyAO:Int, keyEM:Int

' light ranges
Global LR:Float


' ------------------------------------------------------------------------------------------------
' Graphics
' ------------------------------------------------------------------------------------------------

' timing and FPS
Global timer:TTimer = CreateTimer(FPS)
Global period:Int = 1000 / FPS
Global time:Int = MilliSecs() - period
Global F:TFPS = New TFPS  ' untimed FPS counter
Global FT:TFPS = New TFPS ' timed FPS counter

' MaxGUI 3D Window
Global MAXGUI_WINDOW:TGadget = CreateWindow("Shaderdemo (MaxGUI/OpenB3D)", ClientWidth(Desktop()) / 2 - ((SCREEN_WIDTH) / 2), ClientHeight(Desktop()) / 2 - ((SCREEN_HEIGHT) / 2), (SCREEN_WIDTH), SCREEN_HEIGHT, Null, WINDOW_TITLEBAR)
Global MAXGUI_CANVAS:TGadget = CreateCanvas(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, MAXGUI_WINDOW, 0)
SetGadgetLayout(MAXGUI_CANVAS, 1, 1, 1, 1)
ActivateGadget(MAXGUI_CANVAS)
SetGraphics CanvasGraphics(MAXGUI_CANVAS)
Graphics3D(SCREEN_WIDTH, SCREEN_HEIGHT, 32, 0, FPS, -1, True)
EnablePolledInput()


' ------------------------------------------------------------------------------------------------
' Initialization
' ------------------------------------------------------------------------------------------------

' player
Global player:TPivot
Global pivot:TPivot
Global camera:TCamera
InitPlayer()

' light
Global light:TLight
Global lightpivot:TPivot
Global fireball:TLight
Global flicker:Float
Global light2s:TEntity
Global light3s:TEntity
Global light4s:TEntity
Global fireballs:TSprite
InitLights()
AmbientLight FogColor[0] / 2, FogColor[1] / 2, FogColor[2] / 2

' world
Global cube:TMesh
Global col:TMesh
Global walls:TMesh
Global cols:TMesh
Global ground:TMesh
Global ceiling:TMesh
Global sphere:TMesh
InitLevel()

' shader
Global wshader:TShader
Global fshader:TShader
Global cshader:TShader
Global xshader:TShader
Global sshader:TShader
InitShader()

' shininess
EntityShininess ground, 1.0
EntityShininess walls, 1.0
If CREATECEILING Then EntityShininess ceiling, 1.0
If CREATECOLUMNS Then EntityShininess cols, 1.0

' apply shader
If APPLYSHADER Then
	ShadeEntity(sphere, sshader)
	ShadeEntity(ground, fshader)
	ShadeEntity(walls, wshader)
	If CREATECOLUMNS Then ShadeEntity(cols, xshader)
	If CREATECEILING Then ShadeEntity(ceiling, cshader)
EndIf

' movement variables
Global xs:Float
Global ys:Float
Global ya:Float = 0.0
Global px:Float = 0.0
Global yb:Float = 0.0
Global py:Float = 0.0

ResetMouse()

keyAL = 1
keyNM = 1
keyRO = 1
keyME = 1
keyHM = 1
keyAO = 1
keyEM = 1


' ------------------------------------------------------------------------------------------------
' main loop
' ------------------------------------------------------------------------------------------------
While Not KeyDown(KEY_ESCAPE)

	TLight.GetLightValues()
	
	' constant game timing
	Local elapsed:Int
	Repeat elapsed = MilliSecs() - time ; Until elapsed
	For Local k:Int = 1 To elapsed / period ; UpdateGameTimed() ; Next
		
	' user input
	GetUserInput()
		
	' torchlight flicker
	If TL Then
	
		LR = LR1
		PositionEntity light, 0, 0, flicker
		LightColor light, 240, 160, 80
		
	Else
	
		LR = LR1 / 2
		LightColor light, 32, 64, 128
			
	EndIf
	
	' update Camera
	CameraRange camera, Sqr(levelscale), FogRange * 2.0
	CameraClsColor camera, FogColor[0], FogColor[1], FogColor[2]
	
	' torchlight flicker (and emission sinus)
	flicker = ((Sin(timer.Ticks() * 3.0) * Cos(timer.Ticks() * 1.5)) + 1) / 256.0 * levelscale
	
	' fireball movement
	If CREATEFIREBALL Then
	
		m1 = (2 + (2 - Sin(timer.Ticks() * 0.5) * 4.0)) * levelscale
		m2 = 8 * levelscale
		m3 = Cos(timer.Ticks() * 0.5) * levelscale / 2.0
		PositionEntity fireball, m2 + m3, m3, m1 + m3
		
	EndIf

	' set current light range
	LightRange light, LR

	' render
	Wireframe WF
	RenderWorld

	' 2D output
	BeginMax2D()
	
		DrawText("F1 Simple Depth Map......: " + s[DM], 0, 0)
		DrawText("F2 Radial Distance Fog...: " + s[FG], 0, 15)
		DrawText("F3 PBR Shading...........: " + s[pb], 0, 30)
		DrawText("F4 Tonemapping Model.....: " + t[TM], 0, 45)
		DrawText("F5 Camera FOV............: " + FOV + "°", 0, 60)
		DrawText("MB Torchlight............: " + s[TL], 0, 75)
		
		DrawText("1 Albedo Texture........: " + s[keyAL], 0, 100)
		DrawText("2 Normal Texture........: " + s[keyNM], 0, 115)
		DrawText("3 Roughness Texture.....: " + s[keyRO], 0, 130)
		DrawText("4 Metallic Texture......: " + s[keyME], 0, 145)
		DrawText("5 Height Texture........: " + s[keyHM], 0, 160)
		DrawText("6 AO Texture............: " + s[keyAO], 0, 175)
		DrawText("7 Emission Texture......: " + s[keyEM], 0, 190)
				
		DrawText "FPS: " + F.FPS + " (internal: " + FT.FPS + ", average: " + F.averagefps + ")", 0, 210
	
	EndMax2D()
	
	' bugfix: reset light values
	TLight.SetLightValues()
		
	Flip VS
	
	If SCREENSHOT Then

		SetGraphics CanvasGraphics(MAXGUI_CANVAS)
		SavePixmapPNG(GrabPixmap(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT), "FOV" + Int(FOV) + ".png", 0)
		
	EndIf
	
	' update FPS counter
	F.Update()
	
Wend

End


' ------------------------------------------------------------------------------------------------
' timing dependent main loop updates: calculations, movements, rotations
' ------------------------------------------------------------------------------------------------
Function UpdateGameTimed()

	FT.Update()
	time:+period

	Local mx:Int = MouseX()
	Local my:Int = MouseY()
		
	' mouse moved?
	If mx Or my Then

		' get mouse vector
		Local mxs:Float = mx - (SCREEN_WIDTH / 2)
		Local mys:Float = my - (SCREEN_HEIGHT / 2)

		' add to temp var
		xs:+mxs
		ys:+mys

		' center mouse cursor
		MoveMouse(SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2)

		' decrease mouse vector
		xs:*0.3
		ys:*0.3

		' add mouse vector to rotation matrix
		ya:-xs
		px:+ys
		yb:+(ya - yb)
		py:+(px - py)

		' rotate camera and pivot
		RotateEntity camera, py, 0, 0
		RotateEntity pivot, 0, yb, 0

	EndIf

	' player movement
	MoveEntity pivot, (KeyDown(KEY_D) - KeyDown(KEY_A)) * 0.05 * levelscale, 0, (KeyDown(KEY_W) - KeyDown(KEY_S)) * 0.05 * levelscale

	' rotate the colored lights
	TurnEntity lightpivot, 0, 0.5, 0
	TurnEntity sphere, 0, -0.1, 0

End Function


' ------------------------------------------------------------------------------------------------
' get user input
' ------------------------------------------------------------------------------------------------
Function GetUserInput()

	SCREENSHOT = False
	If KeyHit(KEY_TAB) Then VS = 1 - VS
	If KeyHit(KEY_SPACE) Then WF = 1 - WF
	If KeyHit(KEY_BACKSPACE) Then SCREENSHOT = True
	
	If KeyHit(KEY_F1) Then DM:+1 ; DM = DM Mod 2
	If KeyHit(KEY_F2) Then FG:+1 ; FG = FG Mod 2
	If KeyHit(KEY_F3) Then PB:+1 ; PB = PB Mod 2
	If KeyHit(KEY_F4) Then TM:+1 ; TM = TM Mod 7
	
	If KeyHit(KEY_F5) Then
	
		If FOV = 90 Then FOV = 115 Else If FOV = 115 Then FOV = 90
		CameraZoom camera, SetFOV(FOV)
	
	EndIf
	
	If KeyHit(KEY_1) Then keyAL = 1 - keyAL
	If KeyHit(KEY_2) Then keyNM = 1 - keyNM
	If KeyHit(KEY_3) Then keyRO = 1 - keyRO
	If KeyHit(KEY_4) Then keyME = 1 - keyME
	If KeyHit(KEY_5) Then keyHM = 1 - keyHM
	If KeyHit(KEY_6) Then keyAO = 1 - keyAO
	If KeyHit(KEY_7) Then keyEM = 1 - keyEM
	
	If MouseHit(1) Or MouseHit(2) Then TL:+1 ; TL = TL Mod 2
	
End Function


' ------------------------------------------------------------------------------------------------
' create Player
' ------------------------------------------------------------------------------------------------
Function InitPlayer()

	player = CreatePivot()
		
	' main pivot = player
	pivot = CreatePivot(player)
	
	' attach camera to player
	camera = CreateCamera(pivot)
	PositionEntity player, 5 * levelscale, -0.25 * levelscale, 4 * levelscale
	CameraZoom camera, SetFOV(FOV)
	yb = 90
		
	RotateEntity player, 0, -90, 0

End Function


' ------------------------------------------------------------------------------------------------
' create level
' ------------------------------------------------------------------------------------------------
Function InitLights()

	Local lightsprite:TSprite
	Local light2:TLight
	Local light3:TLight
	Local light4:TLight
	Local lighttex:TTexture
	
	' attach light to camera
	light = CreateLight(3, camera)
	'LightConeAngles light, 0, 90
	PositionEntity light, 0, 0, 0
	
	' central pivot for all three lights
	lightpivot = CreatePivot()
	PositionEntity lightpivot, -2.0 * levelscale, -0.8 * levelscale, 6.0 * levelscale
	
	' sprite to visualize the light position
	lightsprite = CreateSprite()
	ScaleSprite lightsprite, 0.05 * levelscale, 0.05 * levelscale
	lighttex = LoadTexture("data/textures/fx/light.tga", 1 + 2)
	EntityTexture lightsprite, lighttex
	TextureBlend lighttex, 3
	EntityFX lightsprite, 1 + 32
	EntityBlend lightsprite, 3
	HideEntity lightsprite

	' additional lights
	If CREATELIGHTS Then
	
		' red light
		light2 = CreateLight(2, lightpivot)
		light2s = CopyEntity(lightsprite, light2)
		EntityColor light2s, 255, 0, 0
		PositionEntity light2, 0.0, 0.25 * levelscale, 0.5 * levelscale
		LightColor light2, 255, 0, 0
		LightRange light2, LR2
		
		' green light
		light3 = CreateLight(2, lightpivot)
		light3s = CopyEntity(lightsprite, light3)
		EntityColor light3s, 0, 255, 0
		PositionEntity light3, 0.0, 0.25 * levelscale, -1.0 * levelscale
		LightColor light3, 0, 255, 0
		LightRange light3, LR3
		
		' blue light
		light4 = CreateLight(2, lightpivot)
		light4s = CopyEntity(lightsprite, light4)
		EntityColor light4s, 0, 0, 255
		PositionEntity light4, 0.0, 0.25 * levelscale, 2.75 * levelscale
		LightColor light4, 0, 0, 255
		LightRange light4, LR4
		
	EndIf
	
	' additional fireball
	If CREATEFIREBALL Then
	
		' fireball
		fireball = CreateLight(3)
		PositionEntity fireball, 0.0, 0.0, 6 * levelscale
		LightColor fireball, 240, 120, 60
		LightRange fireball, LRF
		
		fireballs = CreateSprite(fireball)
		EntityColor fireballs, 240, 120, 60
		ScaleSprite fireballs, 0.125 * levelscale, 0.125 * levelscale
		EntityTexture fireballs, lighttex
		EntityFX fireballs, 1 + 32
		
	EndIf
		
End Function


' ------------------------------------------------------------------------------------------------
' create level
' ------------------------------------------------------------------------------------------------
Function InitLevel()

	' test sphere
	sphere = CreateSphere(32)
	ScaleEntity sphere, levelscale * 0.5, levelscale * 0.5, levelscale * 0.5
	PositionEntity sphere, -2 * levelscale, 0.5 * levelscale, 6 * levelscale
	EntityFX sphere, 32
		
	' world: single cube
	cube = CreateCube()
	ScaleEntity cube, 1.0 * levelscale, 1.0 * levelscale, 1.0 * levelscale
	
	' world: floor
	ground = CreateMesh()
	CreateSurface(ground)
	'EntityFX ground, 32
	
	' world: ceiling
	ceiling = CreateMesh()
	CreateSurface(ceiling)
	'EntityFX ceiling, 32
	
	' create single surface mesh (walls/columns)
	walls = CreateMesh()
	CreateSurface(walls)
	'EntityFX walls, 32
	
	' create single surface mesh (columns)
	If CREATECOLUMNS Then

		' world: single column
		col = CreateCylinder(8)
		ScaleEntity col, 0.2 * levelscale, 1.0 * levelscale, 0.2 * levelscale
	
		cols = CreateMesh()
		CreateSurface(cols)
		'EntityFX cols, 32

	EndIf
	
	' read level data
	RestoreData Level
	For Local y:Int = -10 To 10 Step 2
	
		For Local x:Int = -10 To 10 Step 2
		
			Local val:Int
			ReadData val

			' add ceiling
			If CREATECEILING Then
			
				PositionEntity cube, x * levelscale, 0, y * levelscale
				AddToSurface(cube, ceiling)
				
			EndIf
			
			' add floor
			PositionEntity cube, x * levelscale, 0, y * levelscale
			AddToSurface(cube, ground)
			
			' add walls to single-surface mesh
			If val Then
			
				PositionEntity cube, x * levelscale, 0, y * levelscale
				AddToSurface(cube, walls)
			
			EndIf
			
			' add column template to single-surface mesh
			If x > - 11 And y > - 11 And x < 10 And y < 10 And CREATECOLUMNS Then
			
				' add columns
				PositionEntity col, (x + 1) * levelscale, 0, (y + 1) * levelscale
				AddToSurface(col, cols)
							
			EndIf
	
		Next
		
	Next
	
	' ceiling dimensions
	If CREATECEILING Then
	
		ScaleEntity ceiling, 1, 0, 1
		PositionEntity ceiling, 0, 1 * levelscale, 0
		
	EndIf
	
	' floor dimensions
	ScaleEntity ground, 1, 0, 1
	PositionEntity ground, 0, -1 * levelscale, 0

	' clear template meshes
	FreeEntity cube
	If TMesh(col) Then FreeEntity col
	
End Function


' ------------------------------------------------------------------------------------------------
' initialize all shaders
' ------------------------------------------------------------------------------------------------
Function InitShader()

	' wall shader
	wshader = LoadShader("", "./data/shader/" + sh + ".vert", "./data/shader/" + sh + ".frag")
	SetShaderTexture(wshader, walltex, texturescale_wall)
	
	' floor shader
	fshader = LoadShader("", "./data/shader/" + sh + ".vert", "./data/shader/" + sh + ".frag")
	SetShaderTexture(fshader, floortex, texturescale_floor)

	' ceiling shader
	cshader = LoadShader("", "./data/shader/" + sh + ".vert", "./data/shader/" + sh + ".frag")
	SetShaderTexture(cshader, ceilingtex, texturescale_ceiling)
	
	' columns shader
	xshader = LoadShader("", "./data/shader/" + sh + ".vert", "./data/shader/" + sh + ".frag")
	SetShaderTexture(xshader, columntex, texturescale_columns)
	
	' columns shader
	sshader = LoadShader("", "./data/shader/" + sh + ".vert", "./data/shader/" + sh + ".frag")
	SetShaderTexture(sshader, floortex, texturescale_sphere)

End Function


' ------------------------------------------------------------------------------------------------
' initialize a single shader with different textures but same settings
' ------------------------------------------------------------------------------------------------
Function SetShaderTexture(shader:TShader, tex:String, texscale:Float[])

	' combine the texture filenames
	Local ALfile:String = "./data/textures/" + level + "/" + tex + "." + texformat
	Local NMfile:String = "./data/textures/" + level + "/" + tex + "_Normal." + texformat
	Local ROfile:String = "./data/textures/" + level + "/" + tex + "_Roughness." + texformat
	Local MEfile:String = "./data/textures/" + level + "/" + tex + "_Metallic." + texformat
	Local HMfile:String = "./data/textures/" + level + "/" + tex + "_Height." + texformat
	Local AOfile:String = "./data/textures/" + level + "/" + tex + "_Ambient_Occlusion." + texformat
	Local EMfile:String = "./data/textures/" + level + "/" + tex + "_Emissive." + texformat
	
	' assume there is no texture
	keyAL = 0
	keyNM = 0
	keyRO = 0
	keyME = 0
	keyHM = 0
	keyAO = 0
	keyEM = 0
	
	' check if they exist and set flags if yes
	If(OpenFile(ALfile)) Then keyAL = 1 ; UseInteger(shader, "texAL", keyAL)
	If(OpenFile(NMfile)) Then keyNM = 1 ; UseInteger(shader, "texNM", keyNM)
	If(OpenFile(ROfile)) Then keyRO = 1 ; UseInteger(shader, "texRO", keyRO)
	If(OpenFile(MEfile)) Then keyME = 1 ; UseInteger(shader, "texME", keyME)
	If(OpenFile(HMfile)) Then keyHM = 1 ; UseInteger(shader, "texHM", keyHM)
	If(OpenFile(AOfile)) Then keyAO = 1 ; UseInteger(shader, "texAO", keyAO)
	If(OpenFile(EMfile)) Then keyEM = 1 ; UseInteger(shader, "texEM", keyEM)
			
	' load and pass the existing textures to the shader, set shader flags
	If(keyAL) Then ShaderTexture(shader, LoadTexture(ALfile, 2), "colorMap", 0) Else ShaderTexture(shader, CreateTexture(32, 32), "colorMap", 0)
	If(keyNM) Then ShaderTexture(shader, LoadTexture(NMfile, 2), "normalMap", 1) Else ShaderTexture(shader, CreateTexture(32, 32), "normalMap", 1)
	If(keyRO) Then ShaderTexture(shader, LoadTexture(ROfile, 2), "roughnessMap", 2) Else ShaderTexture(shader, CreateTexture(32, 32), "roughnessMap", 2)
	If(keyME) Then ShaderTexture(shader, LoadTexture(MEfile, 2), "metallicMap", 3) Else ShaderTexture(shader, CreateTexture(32, 32), "metallicMap", 3)
	If(keyHM) Then ShaderTexture(shader, LoadTexture(HMfile, 2), "heightMap", 4) Else ShaderTexture(shader, CreateTexture(32, 32), "heightMap", 4)
	If(keyAO) Then ShaderTexture(shader, LoadTexture(AOfile, 2), "aoMap", 5) Else ShaderTexture(shader, CreateTexture(32, 32), "aoMap", 5)
	If(keyEM) Then ShaderTexture(shader, LoadTexture(EMfile, 2), "emissionMap", 6) Else ShaderTexture(shader, CreateTexture(32, 32), "emissionMap", 6)
	
	' set additional attributes
	SetFloat2(shader, "texscale", texscale[0] * texturescale_multi, texscale[1] * texturescale_multi)
	SetFloat2(shader, "texoffset", texscale[2], texscale[3])
	SetFloat3(shader, "fogColor", FogColor[0] / 255.0, FogColor[1] / 255.0, FogColor[2] / 255.0)
	setFloat(shader, "levelscale", levelscale)
	setFloat(shader, "gamma", gamma)
	SetInteger(shader, "isMetal", keyME)
	
	' set variable runtime attributes
	UseFloat(shader, "lightradius[0].Float", LR)
	UseFloat(shader, "lightradius[1].Float", LR2)
	UseFloat(shader, "lightradius[2].Float", LR3)
	UseFloat(shader, "lightradius[3].Float", LR4)
	UseFloat(shader, "lightradius[4].Float", LRF)
	UseFloat(shader, "flicker", flicker)
	UseFloat(shader, "fogStart", FogStart)
	UseFloat(shader, "fogRange", FogRange)
	UseFloat(shader, "fogDensity", FogDensity)
	UseFloat(shader, "ambFactor", ambFactor)
	UseFloat(shader, "AttA", AttA)
	UseFloat(shader, "AttB", AttB)
	UseFloat(shader, "FOV", FOV)

	' set runtime F-Keys flags
	UseInteger(shader, "flagDM", DM)
	UseInteger(shader, "flagFG", FG)
	UseInteger(shader, "flagPB", pb)
	UseInteger(shader, "flagTM", TM)
	UseInteger(shader, "flagTL", TL)

End Function


' ------------------------------------------------------------------------------------------------
' Reset Mouse to Screen Center
' ------------------------------------------------------------------------------------------------
Function ResetMouse()

	MoveMouse SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2

	MouseXSpeed()
	MouseYSpeed()
	MouseZSpeed()

End Function


' ----------------------------------------------------------------------------
' Convert a given Field of View angle to a Camerazoom value
' ----------------------------------------------------------------------------
Function SetFOV:Float(angle:Float)

	Return 1.0 / Tan(angle / 2.0)

End Function


' ----------------------------------------------------------------------------
' Adds a mesh to another mesh
' ----------------------------------------------------------------------------
Function AddToSurface(source:TMesh, target:TMesh, r:Int = 255, g:Int = 255, b:Int = 255, a:Float = 1.0)

	Local v:Int[3]

	Local s1:TSurface = GetSurface(source, 1)
	Local s2:TSurface = GetSurface(target, CountSurfaces(target))
	
	If CountVertices(s2) + CountVertices(s1) > 32000 Then s2 = CreateSurface(target) ; Print "new surf"
	
	For Local i1:Int = 0 To CountTriangles(s1) - 1

		For Local i2:Int = 0 To 2

			Local t:Int = TriangleVertex(s1, i1, i2)
			TFormPoint VertexX(s1, t), VertexY(s1, t), VertexZ(s1, t), source, target

			v[i2] = AddVertex(s2, TFormedX(), TFormedY(), TFormedZ(), VertexU(s1, t), VertexV(s1, t))
			
			VertexNormal s2, v[i2], VertexNX(s1, t), VertexNY(s1, t), VertexNZ(s1, t)
			VertexColor s2, v[i2], r, g, b, a

		Next

		AddTriangle(s2, v[0], v[1], v[2])

	Next

End Function


' ----------------------------------------------------------------------------
' Creates a Patch
' ----------------------------------------------------------------------------
Function CreatePatch:TMesh(Size:Int)

	Local x:Int, z:Int, v:Float, u:Float, v0:Int, v1:Int, v2:Int, v3:Int
	Local VertexBuffer:Int[129, 129]

	' Create Mesh And surface
	Local mesh:Tmesh = CreateMesh()
	Local surf:Tsurface = CreateSurface(mesh)

	For z = 0 To size

		For x = 0 To size

			' calculate uv coordinates that the texture fits To the tile
			u = x * 1.0 / size
			v = z * 1.0 / size * -1

			' set vertexposition
			VertexBuffer[x, z] = AddVertex (surf, -((size) / 2.0) + x, 0.0, -((size) / 2.0) + z, u, v)
			
			VertexNormal(surf, VertexBuffer[x, z], 0.0, 1.0, 0.0)

		Next
		
	Next

	' set triangles
	For z = 0 To size - 1

		For x = 0 To size - 1

			v0 = VertexBuffer[x, z]
			v1 = VertexBuffer[x + 1, z]
			v2 = VertexBuffer[x + 1, z + 1]
			v3 = VertexBuffer[x, z + 1]

			AddTriangle (surf, v0, v2, v1)
			AddTriangle (surf, v0, v3, v2)

		Next

	Next

	FitMesh(mesh, -0.5, 0, -0.5, 1, 0, 1, 1)

	'UpdateNormals(mesh)

	VertexBuffer = Null

	Return mesh

End Function


' --------------------------------------------------------------------------------------------
' FPS Counter Type
' --------------------------------------------------------------------------------------------
Type TFPS

	Field renders:Int = timer.Ticks()

	Field FPS:Int = 999										' current FPS value
	Field interval:Int = 999
	Field averagefps:Int = 60								' average FPS value
	Field framedrops:Int = 0								' dropped frames
	Field memory:Int = 0									' used memory
	Field oldlooptime:Int = 0
	Field looptime:Int = 0
	Field multi:Float										' main loop multiplicator

	Field old:Int = MilliSecs()
	Field totalfps:Int = 30
	Field Ticks:Int

	' update FPS counter
	Method Update()

		renders:+1

		looptime = MilliSecs() - oldlooptime
		oldlooptime = MilliSecs()

		If MilliSecs() - old >= interval Then

			old = MilliSecs()

			FPS = renders * (1000 / interval)
			renders = 0
			totalfps:+FPS
			Ticks:+1
			
			averagefps = totalfps / Ticks

			memory = GCMemAlloced()

		EndIf

	End Method

End Type


' --------------------------------------------------------------------------------------------
' Leveldata
' --------------------------------------------------------------------------------------------
#Level
DefData 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1
DefData 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1
DefData 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1
DefData 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1
DefData 1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1
DefData 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1
DefData 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1
DefData 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1
DefData 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1
DefData 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1
DefData 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1